#!/usr/bin/env python3
from pwn import*
from sage.all import*
from Crypto.Util.number import *
import json
from tinyec.tinyec.ec import SubGroup, Curve
from Crypto.Util.Padding import pad
from Crypto.Cipher import AES
from json import loads, dumps
from hashlib import sha3_256
from random import choice
from os import urandom

import ast

def recvlineafter(io, lines):
    io.recvuntil(lines)
    rs = io.recvline()
    return rs

local = len(sys.argv) == 1
context.log_level = "debug"
file = "server.py"
io = process(["python", file]) if local else remote(sys.argv[1], int(sys.argv[2]))

xor = lambda a, b: bytes([x^y for x, y in zip(a, b)])
words = b'Enter your command in JSON format:'

n = 0xf1fd178c0b3ad58f10126de8ce42435b53dc67e140d2bf941ffdd459c6d655e1

def to_show_credits(io):
    payload = {"action":"show_credits"}
    payload = json.dumps(payload)
    io.sendlineafter(words ,payload.encode())
    res = io.recvline().decode()
    return res

def to_show_currency(io):
    payload = {"action":"show_currency"}
    payload = json.dumps(payload)
    io.sendlineafter(words ,payload.encode())
    res = ast.literal_eval(io.recvline().decode())["currency"]
    return res
    
def to_get_encrypted_flag(io):
    
    payload = {"action":"get_encrypted_flag"}
    payload = json.dumps(payload)
    io.sendlineafter(words ,payload.encode())
    res = io.recvline().decode()
    return res

def to_get_signatures(io, alea_1, alea_2):
    payload = {"action":"get_signatures", "alea_1":alea_1, "alea_2":alea_2}
    payload = json.dumps(payload)
    io.sendlineafter(words ,payload.encode())
    res = io.recvline().decode().strip()
    return res.replace("\'", "\"").replace("'", '"').replace("(", "[").replace(")", "]")


def to_get_buy_credit(io, owner_proof):
    payload = {"action":"buy_credit", "owner_proofs":owner_proof}
    payload = json.dumps(payload)
    io.sendlineafter(words ,payload.encode())
    res = ast.literal_eval(io.recvline().decode())
    return res

def list_tuple(sigs):
    res = [tuple(sig) for sig in sigs]
    return res

used_credits = 0
sigs, forge_sigs = [], []

def get_new_signatures():
    global used_credits
    global forge_sigs
    global sigs

    new_sigs = json.loads(to_get_signatures(io, "-1", "-2"))["signatures"]

    used_credits +=1 
    
    # forge_sigs = []
    for i, (r, s) in enumerate(new_sigs):
        m = sha3_256(b"this is my lovely loved distributed item " + str(i+10*used_credits).encode()).digest()        
        e = int.from_bytes(sha3_256(m).digest(), "big")
        sigs.append((e, r, s))
        forge_sigs.append((m.hex(), r, n-s))

    return sigs

from tqdm import *

def solve_lll(enc_flag):
    global sigs
    rs, ms, ss = [], [], []
    for sig in sigs:
        m, r, s = sig
        rs.append(r)
        ss.append(s)
        ms.append(m)

    for k_know in trange(2**8):
        k_val = k_know*2**248
        
        B = 2**248
        M = matrix(QQ, len(ms)+2, len(ms)+2)
        order = n
        for i in range(len(ms)):
            M[i,i] = order        

            M[len(ms), i] = mod(rs[i] * inverse(ss[i], order), order) 
            M[len(ms)+1, i] = (-mod(ms[i] * inverse(ss[i], order), order) + k_val) % order 

        M[len(ms), i+1] = QQ(B)/QQ(order)         
        M[len(ms)+1, len(ms)+1] = QQ(B)
        rows = M.LLL()

        for row in rows:    
            if row[-1] == B:
                d = ((QQ(-(row[-2])) * order) / B) % order   
                d1 = int(d)
                d2 = int ( n- d1)    
                flag = decrypt_flag(int(d1).to_bytes(32, "big"), enc_flag)
                flag2 = decrypt_flag(int(d2).to_bytes(32, "big"), enc_flag)
                if b"pwnme" in flag or b"pwnme" in flag2:
                    print(flag)
                    print(flag2)
                    break

def decrypt_flag(key, enc_flag):
    data = ast.literal_eval(enc_flag)
    key = sha3_256(key).digest()[:16]
    enc, iv = data["encrypted_flag"], data["iv"]
    enc = bytes.fromhex(enc)
    iv = bytes.fromhex(iv)
    cipher = AES.new(key, IV=iv, mode=AES.MODE_CBC)
    flag = cipher.decrypt(enc)
    return flag
    # self.used_credit = 0

def main():
    io.recvline()
    enc_flag = to_get_encrypted_flag(io)
    print(f"{enc_flag = }")
    global forge_sigs
    global sigs
    get_new_signatures()
    to_get_buy_credit(io, "")
    for _ in range(4):
        get_new_signatures()
        curr = int(to_show_currency(io))
        to_get_buy_credit(io, forge_sigs[:curr])
        forge_sigs = forge_sigs[curr:]
    get_new_signatures()
    assert len(sigs) == 60
    solve_lll(enc_flag)

    return 


if __name__ == '__main__':
    main()
